import re
import pydash as __


def scream_case(value):
    return __.strings.snake_case(value).upper()


def capitalize(txt=''):
    return txt and txt[0].upper() + txt[1:]


def remove_capital(word):
    return f'{word[0].lower()}{word[1:]}'


def contains_digits(d):
    _digits = re.compile(r'\d')
    return bool(_digits.search(d))


def objPathToTypeParam(fieldPath):
    '''
    from: "brightBeer.name"
    to: "brightBeerName"
    '''
    fieldTags = fieldPath.split('.')
    if len(fieldTags) == 1:
        typeParam = fieldTags[0]
    else:
        typeParam = f'{fieldTags[0]}{capitalize(fieldTags[1])}'
    return typeParam



import os
import click
from gstorm.helpers.Logging import (log, LogLevel)
from gstorm import helpers


@click.command()
@click.option('--src', default=None, help='Graphql schema file to clean')
@click.option('--dst', default=None, help='Graphql schema file path to store the new file')
def cleanup(src, dst):
    [src_path, dst_path] = [os.path.join(
        os.getcwd(), path) for path in [src, dst]]
    helpers.silent_removefile(dst_path)
    with open(src_path, 'r') as src_file:
        with open(dst_path, 'w+') as dst_file:
            for line in src_file:
                is_whitespace = not line.strip()
                if is_whitespace:
                    dst_file.write(line)
                    continue
                line = line.rstrip()  # clean
                line = line.split('#')[0].rstrip()  # remove comment
                # append if not empty
                if line:
                    dst_file.write(line + '\n')
            dst_file.write('\n')
    log(LogLevel.SUCCESS, 'Done')



from __future__ import annotations
import pydash as __
from typing import Optional, List, _GenericAlias, get_type_hints, ForwardRef
from datetime import datetime
from gstorm.helpers.gql import setup_gql
from pprint import pprint
import attr
from dataclasses import dataclass, field
from gstorm.QueryBuilder import QueryBuilder
from gstorm.enums import QueryType
from gstorm.BaseGraphQLType import BaseGraphQLType
from gstorm.helpers.date_helpers import get_local_date, iso8601_to_local_date


def nint(value):
    if value is None:
        return None
    elif type(value) == str:
        return int(value)
    elif type(value) == int:
        return value
    else:
        raise Exception(f"Cannot cast {type(value)} to nullable int")


@attr.s
class GraphQLType(BaseGraphQLType):
    '''Base class for Graphql types,
    has one python's dataclass field per graphql type.
    Plus a __type field with the Type's metadata (opt-out, etc)
    '''
    # __metadata__: str = attr.ib(factory=dict) # attr.ib(default={'opt-out': {}}) # ! MAY BE REQUIRED LATER FOR @opt-out, etc
    __sync__: bool = attr.ib(repr=False, default=False)
    __errors__: List[dict] = attr.ib(repr=False, factory=list)
    id: Optional[int] = attr.ib(
        default=None, converter=nint, metadata={'unique': {}, 'type':int, 'model':None})
    insertedAt: datetime = attr.ib(repr=False, factory=lambda: get_local_date(
        datetime.now()), converter=iso8601_to_local_date, metadata={'readonly': True, 'type':datetime, 'model':None})
    updatedAt: datetime = attr.ib(repr=False, factory=lambda: get_local_date(
        datetime.now()), converter=iso8601_to_local_date, metadata={'readonly': True, 'type':datetime, 'model':None})

    def __attrs_post_init__(self):
        """[summary]
        """
        self._public_keys = self.get_public_fields()

    def __getitem__(self, item):
        if type(item) == str:
            return self.__getattribute__(item)
        elif type(item) == int:
            key = self._public_keys[item]
            return key, self.__getattribute__(key)

    def __setitem__(self, key, value):
        self.__dict__[key] = value

    @classmethod
    def get_field(cls, name):
        return attr.fields_dict(cls).get(name)

    @classmethod
    def get_fields(cls):
        return attr.fields_dict(cls)

    @classmethod
    def get_public_fields(cls):
        return [key for key in cls.get_fields().keys() if not key.startswith('__')]

    @classmethod
    def get_base_fields(cls):
        list_fields = [
            name
            for name, hint in get_type_hints(cls).items()
            if hint != List[GraphQLType]  # if not type List
            # GraphQL types are forward-referenced
            and not type(hint) == ForwardRef
            and not name.startswith('_')
        ]
        return list_fields

    @classmethod
    def get_list_fields(cls):
        list_fields = [
            name
            for name, hint in get_type_hints(cls).items()
            if hint == List[GraphQLType]  # if not type List
        ]
        return list_fields

    @classmethod
    def get_object_fields(cls, typename=None):
        obj_fields = {
            name: hint
            for name, hint in get_type_hints(cls).items()
            if (
                type(hint) == ForwardRef
                # and hint.__forward_arg__ == typename if typename else True # filters by object type
            )
        }
        if typename:
            obj_fields = {
                name: hint
                for name, hint in obj_fields.items()
                if hint.__forward_arg__ == typename
            }
        return list(obj_fields.keys())

    @classmethod
    def get_unique_fields(cls):
        return [
            name
            for name, field in attr.fields_dict(cls).items()
            if 'unique' in field.metadata
        ]

    @classmethod
    def get_field_types(cls):
        return {k: v for k, v in get_type_hints(cls).items() if not k.startswith('_')}

    @classmethod
    def get_metadata(cls, field_name):
        field = attr.fields_dict(cls)[field_name]
        return field.metadata

    def get_unique_identifiers(self):
        return {
            k: self[k]
            for k in self.get_unique_fields()
            if self[k]
        }

    def create(self):
        pass

    def load(self):
        raise NotImplementedError('Child class should implement load method')

    @classmethod
    def query(cls):
        return QueryBuilder(_kind=cls)

    @classmethod
    def query_one(cls):
        return QueryBuilder(cls, QueryType.SINGULAR)

    def is_sync(self):
        return self.__sync__

    def has_errors(self):
        return len(self.__errors__) > 0

    def get_errors(self):
        return self.__errors__

    def update(self, data):
        # Transforms object with selected data
        fields = self.get_fields().keys()
        for key, value in data.items():
            if key in fields:
                self[key] = value
    # ! method aliases
    # query
    q = query
    qm = query
    qp = query  # query plural
    # query_one
    q1 = query_one
    qs = query_one  # query singular



from __future__ import annotations
from typing import get_type_hints, Dict, List, Union, ForwardRef
from dataclasses import dataclass, field
import json
from datetime import datetime
import inflect
from pygqlc import GraphQLClient
import gstorm
from gstorm.enums import QueryType
from gstorm.helpers.str_handling import remove_capital, contains_digits, capitalize
from gstorm.helpers.date_helpers import get_utc_date, get_iso8601_str
from gstorm.helpers.gql import setup_gql
from gstorm.helpers.typing_helpers import is_enum
from gstorm.helpers.gql_schema_helpers import prettify_query
from gstorm.enums import MutationType, MutationMode

'''

# 1. generar mutations
MUT = MutationBuilder(prog)
# 2. Correr mutations
result, errors = gql.mutate(MUT)
# 3. Update de objeto de Planeacion
if errors:
  prog.set_errors(errors)
  prog.set_sync(false)
else:
  prog.set_sync(true)
  prog.id = result.id
if gstorm.validate(prog):
  send_notification('success')
else:
  send_notification('error')

'''


@dataclass
class MutationBuilder():
    data: Union[List[gstorm.GraphQLType], gstorm.GraphQLType] = None
    # _kind: Union[List[gstorm.GraphQLType], gstorm.GraphQLType] = None # ! IMPLICIT FROM DATA
    mode: MutationMode = MutationMode.BATCH
    _mutation: str = ''
    _children: dict = field(default_factory=dict)
    _isChild: bool = False
    _gql: GraphQLClient = field(default_factory=setup_gql)

    def children(self, *args, **kwargs):
        """Appends child objects to mutation, accepts inputs in several formats:
          children(['room', 'inventories']) # ! DEFAULTS TO SAME MUTATION (CREATE, UPDATE)
          children(
            room = MutationType.UNIQUE, # ! MODE, NO CHILDREN IMPLICIT
            inventories = MutationType.CREATE, # ! MODE, NO CHILDREN IMPLICIT
          )
          children({
            'room': MutationType.UNIQUE, # ! MODE, NO CHILDREN IMPLICIT
            'inventories': {
              '__mode': MutationType.CREATE, # ! MODE, CHILDREN EXPLICIT
              'tank': MutationType.UPSERT
            },
          })

        Returns
        -------
        QueryBuilder
          Same instance that called this method for chaining
        """
        if len(args) > 0:
            if type(args[0]) == str:
                self._children = {
                    **self._children,
                    **{args[0]: MutationType.LOCAL}
                }
            elif type(args[0]) == list:
                new_children = {ch: MutationType.LOCAL for ch in args[0]}
                self._children = {
                    **self._children,
                    **new_children
                }
            elif type(args[0]) == dict:
                self._children = {
                    **self._children,
                    **args[0]
                }
        elif kwargs:
            new_children = {ch: _type for ch, _type in kwargs.items()}
            self._children = {**self._children, **kwargs}
        for key in self._children.keys():
            self._children[key]._isChild = True
        return self

    def pretty_compile(self):
        compiled = self.compile()
        if type(compiled) == list:  # sequence of mutations
            return [prettify_query(compiled_item) for compiled_item in compiled]
        else:
            return prettify_query(self.compile())

    def compile(self):
        if type(self.data) == list:
            return self.compile_plural(self.data, self._children)
        else:
            return self.compile_singular(self.data, self._children)

    def compile_plural(self, data, children):

        batch_items = [
            f'create{type(datum).__name__}_{i}: {self.compile_singular(datum, children, batch=True)}'
            for i, datum in enumerate(data)
        ]
        engine = inflect.engine()
        plural_name = engine.plural(remove_capital(type(data[0]).__name__))
        query_name = 'create' + capitalize(plural_name)
        mutation = f'mutation {query_name} {{ ' + " ".join(batch_items) + ' }'

        return mutation

    def compile_singular(self, data, children, batch=False):
        name = f'create{capitalize(type(data).__name__)}'
        if batch:
            _mutation = '$M_NAME $M_PARAMS{ $M_DEF_FIELDS }'
        else:
            _mutation = 'mutation { $M_NAME $M_PARAMS{ $M_DEF_FIELDS } }'
        _m_def_fields = 'successful messages{ field message } result{ $M_FIELDS }'
        fields = [
            k
            for k, hint in get_type_hints(type(data)).items()
            # filter params (only scalar ones)
            if (
                not k.startswith('__')
                and not hint == List[gstorm.GraphQLType]
                and not type(hint) == ForwardRef
            )
        ]
        singular_params = self.compile_singular_params(data)
        child_params = self.compile_child_params(data, children)
        params = singular_params if not child_params else f'{child_params} {singular_params}'
        # compiled result:
        _mutation = _mutation.replace('$M_DEF_FIELDS', _m_def_fields)
        _mutation = _mutation.replace('$M_NAME', name)
        _mutation = _mutation.replace('$M_PARAMS', f'( {params} )')
        _mutation = _mutation.replace('$M_FIELDS', " ".join(fields))
        mutation_singular = _mutation
        # Process list children to include them in the mutation sequence:
        child_plural_mutations = []
        for children_field, _type in children.items():
            if _type == MutationType.CREATE:
                # ! Inject parent into children to include in mutations:
                for child_obj in data[children_field]:
                    # ! Example: Tank.get_object_fields('Room')[0]
                    parent_field = type(child_obj).get_object_fields(
                        type(data).__name__)[0]
                    child_obj[parent_field] = data
                child_plural_mutations.append(self.compile_plural(
                    data[children_field], {parent_field: MutationType.LOCAL}))
        if child_plural_mutations:
            return [mutation_singular, *child_plural_mutations]
        else:
            # ! single mutation, simplest case
            return mutation_singular

    def compile_singular_params(self, data):
        data_type = type(data)
        hints = get_type_hints(data_type)
        params = [
            (k, data[k])
            for k, hint in hints.items()
            # filter params (only scalar ones)
            if (
                not k.startswith('__')
                and not k in ['id', 'insertedAt', 'updatedAt']
                and not hint == List[gstorm.GraphQLType]
                and not type(hint) == ForwardRef
            )
        ]
        return ' '.join([self.gql_param_dumps(k, v) for k, v in params])

    def compile_child_params(self, data, children):
        params = []
        local_children = [
            child
            for child, mtype in children.items()
            if mtype == MutationType.LOCAL
        ]
        for child in local_children:
            child_obj = data[child]
            if not child_obj:
                continue
            # get first non-empty unique identifier:
            _id = next(iter(child_obj.get_unique_identifiers()))
            key = f'{child}{capitalize(_id)}'
            params.append(self.gql_param_dumps(key, child_obj[_id]))
        return ' '.join(params)

    def gql_param_dumps(self, k, v):
        """Converts GraphQL parameter pair (key, value) to it's serialized form
        Examples:
        - {'capacity': 10} -> 'capacity: 10'
        - {'enabled': True} -> 'enabled: true'
        - {'name': 'L101'} -> 'name: "L101"'
        - {'status': 'FINISHED'} -> 'status: FINISHED'

        Parameters
        ----------
        k : param key
            left-hand side id for graphql parameter pair
        v : param value
            right-hand side value for graphql parameter pair
        """
        if type(v) == int:
            return f'{k}: {v}'
        if type(v) == float:
            return f'{k}: {v}'
        if type(v) == str:
            # regular case, this means it's a string, include double quotes
            return f'{k}: "{v}"'
        if type(v) == bool:
            return f'{k}: {["false","true"][v]}'  # short-hand for inline if
        if type(v) == datetime:
            utc_date = get_utc_date(v)
            return f'{k}: "{get_iso8601_str(utc_date)}"'
        if type(v) == dict:
            quote = '"'
            scape = '\\'
            return f'{k}: "{json.dumps(v).replace(quote, scape + quote)}"'
        if is_enum(v):
            return f'{k}: {v.value}'
        if v is None:
            return f'{k}: null'

    def run(self):

        _mutation = self.compile()
        if type(_mutation) == str:
            # ! SINGLE MUTATION
            data, errors = self._gql.mutate(_mutation)
            if type(self.data) == list:
                # ! BATCH MUTATION
                full_response = {'result': [],
                                 'messages': [], 'successful': []}
                for i, item in enumerate(self.data):
                    label = f'create{type(item).__name__}_{i}'
                    # data for this specific item:
                    item_response = data.get(label)
                    # append to full response:
                    full_response['successful'].append(
                        item_response.get('successful', False))
                    full_response['messages'].extend(
                        item_response.get('messages', []))
                    full_response['result'].append(
                        item_response.get('result', {}))
                    item.__errors__ = item_response['messages'] if len(
                        item_response.get('messages', [])) > 0 else []
                    if 'result' in item_response:
                        item.update(item_response['result'])
                        item.__sync__ = True
                    else:
                        item.__errors__.extend(
                            [{'message': 'No data for this item'}])
                return full_response
            else:
                # ! SIMPLE MUTATION
                self.data.__errors__ = errors
                if not errors:
                    # rebuilds object adding data from DB
                    self.data.update(data['result'])
                    self.data.__sync__ = True
                return data
        elif type(_mutation) == list:
            _mutations = _mutation
            # ! SEQUENTIAL MUTATION
            sequence = [self.data]
            sequence.extend([
                self.data[child]
                for child, _type in self._children.items()
                if _type == MutationType.CREATE
            ])
            full_response = {'result': [], 'messages': [], 'successful': []}
            for mutation, seq_obj in zip(_mutations, sequence):
                # ! Run every mutation sequentially
                data, errors = self._gql.mutate(mutation)
                if type(seq_obj) == list:
                    # ! SEQUENCE - BATCH MUTATION ITEM
                    for i, item in enumerate(seq_obj):
                        label = f'create{type(item).__name__}_{i}'
                        # data for this specific item:
                        item_response = data.get(label)
                        # append to full response:
                        full_response['successful'].append(
                            item_response.get('successful', False))
                        full_response['messages'].extend(
                            item_response.get('messages', []))
                        full_response['result'].append(
                            item_response.get('result', {}))
                        item.__errors__ = item_response['messages'] if len(
                            item_response.get('messages', [])) > 0 else []
                        if 'result' in item_response:
                            item.update(item_response['result'])
                            item.__sync__ = True
                        else:
                            item.__errors__.extend(
                                [{'message': 'No data for this item'}])
                else:
                    # ! SEQUENCE - SINGULAR MUTATION ITEM
                    seq_obj.__errors__ = errors
                    if not errors:
                        # rebuilds object adding data from DB
                        seq_obj.update(data['result'])
                        seq_obj.__sync__ = True
            return full_response
        else:
            raise Exception(
                'Compiled mutation should be either a string or a list of strings')



import os
from gstorm.helpers.Logging import log, LogLevel


def inmemory_cleanup(src):
    src_path = os.path.join(os.getcwd(), src)
    output = []
    with open(src_path, 'r') as src_file:
        for line in src_file:
            is_whitespace = not line.strip()
            if is_whitespace:
                output.append(line)
                continue
            line = line.rstrip()  # clean
            line = line.split('#')[0].rstrip()  # remove comment
            # append if not empty
            if line:
                output.append(line + '\n')
        output.append('\n')
    return output


def scalar_map(field_typename):
    return {
        'Boolean': 'Boolean',
        'Integer': 'Int',
        'Float': 'Float',
        'String': 'String',
        'Text': 'Text',
        'Json': 'map',
        'Jsonb': 'map',
        'Date': 'Date',
        'Time': 'Time',
        'Datetime': 'DateTime'
    }[field_typename]


def get_field_kind(field_typename):
    scalar_types = [
        'Boolean',
        'Integer',
        'Float',
        'String',
        'Text',
        'Json',
        'Jsonb',
        'Date',
        'Time',
        'Datetime'
    ]
    if field_typename in scalar_types:
        return 'SCALAR'
    else:
        return 'OBJECT'


'''
Syntactic analizer for a graphql schema

things to consider:

# ! 1. general structure of a type
type TYPE_NAME @opts {
  ...
}
# ! 2. general structure of an enum
enum ENUM_TYPE @opts {
  ...
}
# ! 3. All basic types
type MyType {
  enabled: Boolean
  quantity: Int
  weight: Float
  name: String
  description: Text
  config: Json
  shipAt: Datetime
}
# ! 4. Complex type relations
# ? 4.1 ComplexType1.ListOf.ComplexType2 (Cannot be a list of BasicType)
# ? 4.2 ComplexType1.hasOne.ComplexType2 (Cannot be a hasOne of BasicType)
# ? 4.3 ComplexType1.ComplexType2 (ComplexType2 Must exist)
type User {
  name: String!
  address: Address @has_one
  posts: [Blogpost]
}

type Address {
  street: String
  city: String
  number: Integer
}

type Blogpost {
  author: User
  title: String @unique
}

type UserSkill {
  user: User
  skill: Skill
  level: Integer!
}

type Skill {
  name: String
}
'''


def gql_schema_parse(schema_file_lines):
    state = 'TYPE_SEARCH'
    schema = {}  # dict of typenames
    # context
    current_type = None
    current_field = None
    for line in schema_file_lines:
        line = line.strip()
        tokens = line.split()
        if (state == 'TYPE_SEARCH'):
            if len(tokens) == 0:
                continue
            if tokens[0] == 'type':
                state = 'PARSING_TYPE'
                current_type = {}
                current_type['name'] = tokens[1]
                current_type['kind'] = 'TYPE'
                current_type['fields'] = []
            elif tokens[0] == 'enum':
                state = 'PARSING_ENUM'
                current_type = {}
                current_type['name'] = tokens[1]
                current_type['kind'] = 'ENUM'
                current_type['enumValues'] = []
            else:
                continue
        elif (state == 'PARSING_TYPE'):
            if len(tokens) == 0:
                # blank line inside the type
                continue
            elif tokens[0] == '}':
                # end of type definition
                state = 'TYPE_SEARCH'
                schema[current_type['name']] = current_type
                current_type = None
            else:
                # type body (should be a field definition)
                field_name = tokens[0].strip(':')
                field_is_required = '!' in tokens[1]
                field_is_list = '[' in tokens[1] and ']' in tokens[1]
                field_typename = tokens[1].strip('!').strip('[]').strip('!')
                field_kind = get_field_kind(field_typename)
                field_has_one = False if len(
                    tokens) < 3 else '@has_one' == tokens[2]
                field_is_unique = False if len(
                    tokens) < 3 else '@unique' == tokens[2]
                if len(tokens) < 3 or not '@default' in tokens[2]:
                    field_default = None
                else:
                    field_default = tokens[3][:-1]
                current_field = {}
                # current_field['required'] = field_is_required
                # current_field['is_unique'] = field_is_unique
                current_field['unique'] = field_is_unique
                current_field['has_one'] = field_has_one
                current_field['default'] = field_default
                current_field['name'] = field_name
                current_field['type'] = {
                    'name': field_typename if field_kind == 'OBJECT' else scalar_map(field_typename),
                    'kind': 'LIST' if field_is_list else field_kind
                }
                current_type['fields'].append(current_field)
                current_field = None
        elif (state == 'PARSING_ENUM'):
            if len(tokens) == 0:
                # blank line inside the enum
                continue
            elif tokens[0] == '}':
                # end of enum definition
                state = 'TYPE_SEARCH'
                schema[current_type['name']] = current_type
                current_type = None
            else:
                # enum body (should be a EnumValue definition)
                current_type['enumValues'].append(tokens[0].strip())
    # we now know every schema type, we need to update the fields type kind to ENUM if needed:
    for typename, item in schema.items():
        if item['kind'] == 'ENUM':
            continue
        for index, field in enumerate(item['fields']):
            if not field['type']['name'] in schema:
                # may be an error, but leave it to the validation function
                continue
            if (
                field['type']['kind'] == 'OBJECT'
                and schema[field['type']['name']]['kind'] == 'ENUM'
            ):
                schema[typename]['fields'][index]['type']['kind'] = 'ENUM'
    return schema


def load_schema_from_file(src):
    try:
        clean_schema_lines = inmemory_cleanup(src)
        return gql_schema_parse(clean_schema_lines)
    except FileNotFoundError:
        log(LogLevel.ERROR, f'File {src} not found')
        return


def prettify_query(query):
    tabs = 0
    pretty_query = ''
    parsing_string = False
    inside_brackets = False
    for i, ch in enumerate(query):
        # get context
        prev_ch = query[i-1] if i > 0 else ''
        next_ch = query[i+1] if i < len(query) - 1 else ''
        if not inside_brackets and ch == '{':
            inside_brackets = True
        elif not inside_brackets:
            pretty_query += ch
            continue
        if not parsing_string and ch == '"':
            parsing_string = True
            pretty_query += ch
            continue
        if parsing_string and ch != '"':
            pretty_query += ch
            continue
        if parsing_string and ch == '"':
            parsing_string = False
            pretty_query += ch
            continue
        tabs += 1 if ch in ['{', '('] else (-1 if next_ch in ['}', ')'] else 0)
        if ch != ' ' or prev_ch == ':':
            pretty_query += ch
            continue
        if next_ch in ['{', '(']:
            pretty_query += ch  # add whitespace normally
        elif next_ch in ['}', ')']:
            pretty_query += f"\n{'  ' * tabs}"
        else:
            pretty_query += f"\n{'  ' * tabs}"
    return pretty_query



import os
from pprint import pformat
import click
from gstorm.helpers.Logging import log, LogLevel
from gstorm.helpers.gql_schema_helpers import load_schema_from_file


def gql_schema_validate(schema):
    errors = []
    for typename, item in schema.items():
        log(LogLevel.INFO, f'\nvalidating type {typename}')
        # ! Style validation
        if typename[0].islower() or '_' in typename:
            # * No lowercasing Types
            errors.append({
                'type': item['name'],
                'field': field['name'],
                'message': f'Typename "{typename}" should be in CapitalCase'
            })
        if item['kind'] == 'TYPE':
            # ! validate every field
            for field in item['fields']:
                log(LogLevel.INFO, f'\tvalidating field {field["name"]}')
                # ! Style validation
                if field['name'][0].isupper():
                    # * NO Capitalizing fields
                    errors.append({
                        'type': item['name'],
                        'field': field['name'],
                        'message': f'Field "{field["name"]}" should not be capitalized'
                    })
                elif '_' in field['name']:
                    errors.append({
                        'type': item['name'],
                        'field': field['name'],
                        'message': f'Field "{field["name"]}" should be in camelCase'
                    })
                # ! Schema validation
                if field['type']['kind'] == 'SCALAR':
                    # u all right, fam
                    continue
                elif field['type']['kind'] == 'LIST':
                    # first of all, this type exists?
                    if not field['type']['name'] in schema:
                        errors.append({
                            'type': item['name'],
                            'field': field['name'],
                            'message': f'LIST type "{field["type"]["name"]}" does not exist'
                        })
                        continue
                    # validate List Referring type has a field with corresponding type
                    referred_type = schema[field['type']['name']]
                    referred_fields = [
                        ref_field for ref_field in referred_type['fields']
                        if ref_field['type']['name'] == typename
                        and not ref_field['has_one']
                    ]
                    if len(referred_fields) == 0:
                        errors.append({
                            'type': item['name'],
                            'field': field['name'],
                            'message': 'LIST Has no corresponding object in referred type'
                        })
                    # elif len(referred_fields) > 1:
                    #   errors.append({
                    #     'type': item['name'],
                    #     'field': field['name'],
                    #     'message': 'LIST has too many referred types (Should be only one)'
                    #   })
                elif field['type']['kind'] == 'OBJECT' and field['has_one']:
                    # first of all, this type exists?
                    if not field['type']['name'] in schema:
                        errors.append({
                            'type': item['name'],
                            'field': field['name'],
                            'message': f'OBJECT type "{field["type"]["name"]}" does not exist'
                        })
                        continue
                    # validate Object Referring type has a field with corresponding type
                    referred_type = schema[field['type']['name']]
                    referred_fields = [
                        ref_field for ref_field in referred_type['fields']
                        if ref_field['type']['name'] == typename
                        and not ref_field['has_one']
                    ]
                    if len(referred_fields) == 0:
                        errors.append({
                            'type': item['name'],
                            'field': field['name'],
                            'message': 'HAS_ONE has no corresponding object in referred type'
                        })
                    # elif len(referred_fields) > 1:
                    #   errors.append({
                    #     'type': item['name'],
                    #     'field': field['name'],
                    #     'message': 'HAS_ONE has too many referred types (Should be only one)'
                    #   })
                elif field['type']['kind'] == 'OBJECT' and not field['has_one']:
                    # first of all, this type exists?
                    if not field['type']['name'] in schema:
                        errors.append({
                            'type': item['name'],
                            'field': field['name'],
                            'message': f'Object field type "{field["type"]["name"]}" does not exist'
                        })
                        continue
                    # validate List Referring type has a field with corresponding type
                    referred_type = schema[field['type']['name']]
                    referred_fields_has_one = [
                        ref_field for ref_field in referred_type['fields']
                        if ref_field['type']['name'] == typename
                        and ref_field['has_one']
                    ]
                    referred_fields_list = [
                        ref_field for ref_field in referred_type['fields']
                        if ref_field['type']['name'] == typename
                        and ref_field['type']['kind'] == 'LIST'
                    ]
                    if (
                        len(referred_fields_has_one) > 0
                        and len(referred_fields_list) > 0
                    ):
                        errors.append({
                            'type': item['name'],
                            'field': field['name'],
                            'message': 'OBJECT reference is both HAS_ONE and LIST in referred type'
                        })
                    elif (
                        len(referred_fields_has_one) == 0
                        and len(referred_fields_list) == 0
                    ):
                        errors.append({
                            'type': item['name'],
                            'field': field['name'],
                            'message': f'field {field["name"]} has no corresponding LIST or HAS_ONE'
                        })
        elif item['kind'] == 'ENUM':
            # ! validate Enumeration usage
            schema_object_types = {tn2: i2 for tn2,
                                   i2 in schema.items() if i2['kind'] == 'TYPE'}
            referring_types = [
                1 for typename2, item2 in schema_object_types.items()
                if len([
                    1 for field in item2['fields']
                    if field['type']['name'] == item['name']
                ]) > 0
            ]
            if len(referring_types) == 0:
                errors.append({
                    'type': item['name'],
                    'field': '',
                    'message': 'ENUM does not have any referring type (should be at least one)'
                })
    return errors


@click.command()
@click.option('--src', default=None, help='Graphql schema file to analyze')
def validate_schema(src: str):
    """Validates graphql schema relations

    Parameters
    ----------
    src : str
      relative path to schema file (.graphql)
      ex: "gstorm-cli validate-schema --src=graphql/doc/schema.graphql"
    """
    # ! THIS
    schema = load_schema_from_file(src)
    from pprint import pprint
    pprint(schema)
    # ! UNTIL HERE
    validation_errors = gql_schema_validate(schema)
    if len(validation_errors) > 0:
        log(LogLevel.ERROR,
            f'\n\n{len(validation_errors)} ERROR FOUND INSPECTING THE SCHEMA:')
        log(LogLevel.ERROR, pformat(validation_errors))
        log(LogLevel.ERROR, '\n\n')
    else:
        log(LogLevel.SUCCESS, '\n\nNo errors found in the schema :)\n\n')



import os
import pathlib
import click
from .utils import one_to_one_map, imports, example, default_map
from gstorm.helpers.gql_schema_helpers import load_schema_from_file
from gstorm.helpers.Logging import log, LogLevel
from gstorm.helpers.os_helpers import query_yes_no
from gstorm.helpers.str_handling import scream_case


def _build_classes(schema, dir_to_save, mpath):

    for class_name, attributes in schema.items():
        path_and_file = os.path.join(dir_to_save, class_name + ".py")
        log(LogLevel.WARNING, f'Building class file for {class_name} model')
        # Extract class attributes for enumeration
        cls_attrs_buffer = ''
        with open(path_and_file, 'w') as file:
            # * Clases GraphQLType
            if attributes['kind'] == 'TYPE':
                cls_attrs = [
                    scream_case(field['name'])
                    for field in attributes['fields']
                    if field['type']['kind'] == 'SCALAR'
                ]
                cls_attrs = ['ID'] + \
                    sorted(cls_attrs) + ['INSERTED_AT', 'UPDATED_AT']
                # build ClassNameAttrs class (Enum):
                cls_attrs_buffer = f'\nclass {class_name}Attrs(Enum):\n'
                for attr in cls_attrs:
                    cls_attrs_buffer += f"\t{attr} = '{attr}'\n"
                file.write(imports)
                buffer_written = f"\nmpath = '{mpath}'\n\n@attr.s\nclass {class_name}(GraphQLType):\n"
                for field in attributes['fields']:
                    name = field['name']
                    kind = field['type']['kind']
                    temp = field['type']['name']
                    default_field = field['default']
                    has_one = field['has_one']
                    unique = field['unique']
                    metadata = "metadata={{{0}{1}"\
                        .format("'unique':True," if unique else '',
                                "'has_one':True," if has_one else '')
                    if default_field == None:
                        type_name, default = (
                            temp, "None") if temp not in one_to_one_map else one_to_one_map[temp]
                    else:
                        type_name, default = default_map(temp, default_field)

                    if kind == "OBJECT" and temp != 'Int':
                        metadata += f"'type':GraphQLType,'model':'{type_name}'}}"
                        buffer_written += \
                            f"    {name}: '{type_name}' = attr.ib(default={default}, repr=gql_repr, converter=convert_to(mpath,'{type_name}'), {metadata})\n"
                    elif kind == "LIST":
                        metadata += f"'type':list,'model':'{type_name}'}}"
                        buffer_written += \
                            f"    {name}: List[GraphQLType] = attr.ib(factory=list, repr=gql_list_repr, converter=list_convert_to(mpath,'{type_name}'), {metadata})\n"
                    elif kind == "ENUM":
                        metadata += f"'type':Enum,'model':'{type_name}'}}"
                        file.write(f"from .{type_name} import {type_name}\n")
                        buffer_written += \
                            f"    {name}: {type_name} = attr.ib(default='{schema[type_name]['enumValues'][0]}', converter=convert_to(mpath,'{type_name}'), {metadata})\n"
                    elif type_name == 'datetime':
                        metadata += f"'type':{type_name},'model':None}}"
                        buffer_written += \
                            f"    {name}: {type_name} = attr.ib(default={default}, converter=iso8601_to_local_date, {metadata})\n"
                    else:
                        metadata += f"'type':{type_name},'model':None}}"
                        buffer_written += \
                            f"    {name}: {type_name} = attr.ib(default={default}, {metadata})\n"

            # * Clases Enum
            elif attributes['kind'] == 'ENUM':
                file.write("from enum import Enum")
                buffer_written = f"\n\nclass {class_name}(Enum):\n"
                for enum in attributes['enumValues']:
                    buffer_written += f"    {enum} = '{enum}'\n"
            if (cls_attrs_buffer):
                file.write(cls_attrs_buffer)
            file.write(buffer_written)

    # Creacion del archivo __init__.py
    path_and_file = os.path.join(dir_to_save, "__init__.py")
    with open(path_and_file, 'w') as file:
        models, enums = [], []
        file.write('\n# Models and Attribute enums\n')
        for class_name, attributes in schema.items():
            if attributes['kind'] != 'TYPE':
                continue
            file.write(
                f"from .{class_name} import {class_name}, {class_name}Attrs\n")
            models.append(class_name)
        file.write('\n# Enumerations\n')
        for class_name, attributes in schema.items():
            if attributes['kind'] != 'ENUM':
                continue
            file.write(f"from .{class_name} import {class_name}\n")
            enums.append(class_name)
        
        # Global lists
        comma = '",\n    "'
        file.write(f'''
__models__ = [
    "{comma.join([class_name for class_name in models])}"
]
__enums__ = [
    "{comma.join([class_name for class_name in enums])}"
]
''')


@click.command()
@click.option('--src', default=None, help='Graphql schema file to analyze')
@click.option('--output', default=None, help='output directory for the GQL model files')
def build_classes(src: str, output: str):
    # from "tests/models" or "tests/models/" to "tests.models"
    mpath = output.rstrip('/').replace('/', '.')
    output = os.path.join(os.getcwd(), output)
    out_path_exists = os.path.isdir(output)
    if not out_path_exists:
        log(LogLevel.WARNING,
            'Output directory not found, this command will create a directory with files in:')
        log(LogLevel.WARNING, f'{output}')
        answer = query_yes_no("Do you want to proceed?",
                              "no", LogLevel.WARNING)
        if not answer:
            log(LogLevel.INFO, 'Exiting without building any class.')
            return
        pathlib.Path(output).mkdir(parents=True, exist_ok=True)
    schema = load_schema_from_file(src)
    if (not schema):
        raise Exception("Error parsing schema")
    _build_classes(schema, output, mpath)
    log(LogLevel.SUCCESS, 'Done')
